import datetime
from typing import Any, Final

import astropy.units as u
from astropy.units import astrophys, cgs, core, misc, si
from pydantic import (
    UUID1,
    UUID3,
    UUID4,
    UUID5,
    UUID6,
    UUID7,
    UUID8,
    AmqpDsn,
    AnyHttpUrl,
    AnyUrl,
    AnyWebsocketUrl,
    Base64Bytes,
    Base64Str,
    Base64UrlBytes,
    Base64UrlStr,
    ClickHouseDsn,
    CockroachDsn,
    DirectoryPath,
    EmailStr,
    FilePath,
    FileUrl,
    FiniteFloat,
    FtpUrl,
    GetCoreSchemaHandler,
    HttpUrl,
    IPvAnyAddress,
    KafkaDsn,
    MariaDBDsn,
    MongoDsn,
    MySQLDsn,
    NameEmail,
    NatsDsn,
    NegativeFloat,
    NegativeInt,
    NonNegativeFloat,
    NonNegativeInt,
    NonPositiveFloat,
    NonPositiveInt,
    PositiveFloat,
    PositiveInt,
    PostgresDsn,
    RedisDsn,
    SnowflakeDsn,
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
    WebsocketUrl,
)
from pydantic_core import core_schema

# --- Astropy Physical Types Logic ---


def create_quantity_type(name: str, target_physical_type: str):
    class QuantityType(str):
        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Any, handler: GetCoreSchemaHandler
        ) -> core_schema.CoreSchema:
            return core_schema.no_info_after_validator_function(
                cls.validate,
                core_schema.str_schema(),
            )

        @classmethod
        def validate(cls, v: str) -> str:
            try:
                q = u.Quantity(v)
                # Ensure we are comparing PhysicalType objects to handle synonyms
                assert q.unit is not None, "Quantity unit cannot be None"
                pt = q.unit.physical_type
                target_pt = u.get_physical_type(target_physical_type)

                if pt != target_pt:
                    raise ValueError(
                        f"Physical type mismatch: expected '{target_physical_type}' ({target_pt}), got '{pt}'"
                    )
            except Exception as e:
                # Wrap any astropy parsing error or our custom mismatch error
                raise ValueError(f"Invalid quantity for type '{name}': {e}") from e
            return v

    QuantityType.__name__ = name
    return QuantityType


# List provided by user
_units_and_physical_types: Final[list[tuple[core.UnitBase, str | set[str]]]] = [
    (core.dimensionless_unscaled, "dimensionless"),
    (si.m, "length"),
    (si.m**2, "area"),
    (si.m**3, "volume"),
    (si.s, "time"),
    (si.rad, "angle"),
    (si.sr, "solid angle"),
    (si.m / si.s, {"speed", "velocity"}),
    (si.m / si.s**2, "acceleration"),
    (si.Hz, "frequency"),
    (si.g, "mass"),
    (si.mol, "amount of substance"),
    (si.K, "temperature"),
    (si.W * si.m**-1 * si.K**-1, "thermal conductivity"),
    (si.J * si.K**-1, {"heat capacity", "entropy"}),
    (si.J * si.K**-1 * si.kg**-1, {"specific heat capacity", "specific entropy"}),
    (si.N, "force"),
    (si.J, {"energy", "work", "torque"}),
    (si.J * si.m**-2 * si.s**-1, {"energy flux", "irradiance"}),
    (si.Pa, {"pressure", "energy density", "stress"}),
    (si.W, {"power", "radiant flux"}),
    (si.kg * si.m**-3, "mass density"),
    (si.m**3 / si.kg, "specific volume"),
    (si.mol / si.m**3, "molar concentration"),
    (si.m**3 / si.mol, "molar volume"),
    (si.kg * si.m / si.s, {"momentum", "impulse"}),
    (si.kg * si.m**2 / si.s, {"angular momentum", "action"}),
    (si.rad / si.s, {"angular speed", "angular velocity", "angular frequency"}),
    (si.rad / si.s**2, "angular acceleration"),
    (si.rad / si.m, "plate scale"),
    (si.g / (si.m * si.s), "dynamic viscosity"),
    (si.m**2 / si.s, {"diffusivity", "kinematic viscosity"}),
    (si.m**-1, "wavenumber"),
    (si.m**-2, "column density"),
    (si.A, "electrical current"),
    (si.C, "electrical charge"),
    (si.V, "electrical potential"),
    (si.Ohm, {"electrical resistance", "electrical impedance", "electrical reactance"}),
    (si.Ohm * si.m, "electrical resistivity"),
    (si.S, "electrical conductance"),
    (si.S / si.m, "electrical conductivity"),
    (si.F, "electrical capacitance"),
    (si.C * si.m, "electrical dipole moment"),
    (si.A / si.m**2, "electrical current density"),
    (si.V / si.m, "electrical field strength"),
    (
        si.C / si.m**2,
        {"electrical flux density", "surface charge density", "polarization density"},
    ),
    (si.C / si.m**3, "electrical charge density"),
    (si.F / si.m, "permittivity"),
    (si.Wb, "magnetic flux"),
    (si.Wb**2, "magnetic helicity"),
    (si.T, "magnetic flux density"),
    (si.A / si.m, "magnetic field strength"),
    (si.m**2 * si.A, "magnetic moment"),
    (si.H / si.m, {"electromagnetic field strength", "permeability"}),
    (si.H, "inductance"),
    (si.cd, "luminous intensity"),
    (si.lm, "luminous flux"),
    (si.lx, {"luminous emittance", "illuminance"}),
    (si.W / si.sr, "radiant intensity"),
    (si.cd / si.m**2, "luminance"),
    (si.m**-3 * si.s**-1, "volumetric rate"),
    (astrophys.Jy, "spectral flux density"),
    (astrophys.Jy / si.sr, "surface brightness"),
    (si.W * si.m**2 * si.Hz**-1, "surface tension"),
    (si.J * si.m**-3 * si.s**-1, {"spectral flux density wav", "power density"}),
    (si.J * si.m**-3 * si.s**-1 * si.sr**-1, "surface brightness wav"),
    (astrophys.photon / si.Hz / si.cm**2 / si.s, "photon flux density"),
    (astrophys.photon / si.AA / si.cm**2 / si.s, "photon flux density wav"),
    (astrophys.photon / si.Hz / si.cm**2 / si.s / si.sr, "photon surface brightness"),
    (
        astrophys.photon / si.AA / si.cm**2 / si.s / si.sr,
        "photon surface brightness wav",
    ),
    (astrophys.R, "photon flux"),
    (misc.bit, "data quantity"),
    (misc.bit / si.s, "bandwidth"),
    (cgs.Franklin, "electrical charge (ESU)"),
    (cgs.statampere, "electrical current (ESU)"),
    (cgs.Biot, "electrical current (EMU)"),
    (cgs.abcoulomb, "electrical charge (EMU)"),
    (si.m * si.s**-3, {"jerk", "jolt"}),
    (si.m * si.s**-4, {"snap", "jounce"}),
    (si.m * si.s**-5, "crackle"),
    (si.m * si.s**-6, {"pop", "pounce"}),
    (si.K / si.m, "temperature gradient"),
    (si.J / si.kg, {"specific energy", "dose of ionizing radiation"}),
    (si.mol * si.m**-3 * si.s**-1, "reaction rate"),
    (si.kg * si.m**2, "moment of inertia"),
    (si.mol / si.s, "catalytic activity"),
    (si.J * si.K**-1 * si.mol**-1, "molar heat capacity"),
    (si.mol / si.kg, "molality"),
    (si.m * si.s, "absement"),
    (si.m * si.s**2, "absity"),
    (si.m**3 / si.s, "volumetric flow rate"),
    (si.s**-2, "frequency drift"),
    (si.Pa**-1, "compressibility"),
    (astrophys.electron * si.m**-3, "electron density"),
    (astrophys.electron * si.m**-2 * si.s**-1, "electron flux"),
    (si.kg / si.m**2, "surface mass density"),
    (si.W / si.m**2 / si.sr, "radiance"),
    (si.J / si.mol, "chemical potential"),
    (si.kg / si.m, "linear density"),
    (si.H**-1, "magnetic reluctance"),
    (si.W / si.K, "thermal conductance"),
    (si.K / si.W, "thermal resistance"),
    (si.K * si.m / si.W, "thermal resistivity"),
    (si.N / si.s, "yank"),
    (si.S * si.m**2 / si.mol, "molar conductivity"),
    (si.m**2 / si.V / si.s, "electrical mobility"),
    (si.lumen / si.W, "luminous efficacy"),
    (si.m**2 / si.kg, {"opacity", "mass attenuation coefficient"}),
    (si.kg * si.m**-2 * si.s**-1, {"mass flux", "momentum density"}),
    (si.m**-3, "number density"),
    (si.m**-2 * si.s**-1, "particle flux"),
]


# Generate map of astropy types
ASTROPY_TYPES = {}
for _unit, names in _units_and_physical_types:
    if isinstance(names, str):
        names_set = {names}
    else:
        names_set = names

    for name in names_set:
        ASTROPY_TYPES[name] = create_quantity_type(name, name)

# --- Standard Pydantic Types ---

STANDARD_TYPES = {
    "str": str,
    "string": str,
    "date": datetime.date,
    "datetime": datetime.datetime,
    "clocktime": datetime.time,  # Note: 'time' renamed to 'clocktime' to avoid conflict with astropy 'time' type
    "int": int,
    "float": float,
    "bool": bool,
    "path": str,
    "url": str,
    "any": Any,
    "markdown": str,
    "StrictBool": StrictBool,
    "PositiveInt": PositiveInt,
    "NegativeInt": NegativeInt,
    "NonPositiveInt": NonPositiveInt,
    "NonNegativeInt": NonNegativeInt,
    "StrictInt": StrictInt,
    "PositiveFloat": PositiveFloat,
    "NegativeFloat": NegativeFloat,
    "NonPositiveFloat": NonPositiveFloat,
    "NonNegativeFloat": NonNegativeFloat,
    "StrictFloat": StrictFloat,
    "FiniteFloat": FiniteFloat,
    "StrictStr": StrictStr,
    "UUID1": UUID1,
    "UUID3": UUID3,
    "UUID4": UUID4,
    "UUID5": UUID5,
    "UUID6": UUID6,
    "UUID7": UUID7,
    "UUID8": UUID8,
    "FilePath": FilePath,
    "DirectoryPath": DirectoryPath,
    "Base64Bytes": Base64Bytes,
    "Base64Str": Base64Str,
    "Base64UrlBytes": Base64UrlBytes,
    "Base64UrlStr": Base64UrlStr,
    "AnyUrl": AnyUrl,
    "AnyHttpUrl": AnyHttpUrl,
    "HttpUrl": HttpUrl,
    "AnyWebsocketUrl": AnyWebsocketUrl,
    "WebsocketUrl": WebsocketUrl,
    "FileUrl": FileUrl,
    "FtpUrl": FtpUrl,
    "PostgresDsn": PostgresDsn,
    "CockroachDsn": CockroachDsn,
    "AmqpDsn": AmqpDsn,
    "RedisDsn": RedisDsn,
    "MongoDsn": MongoDsn,
    "KafkaDsn": KafkaDsn,
    "NatsDsn": NatsDsn,
    "MySQLDsn": MySQLDsn,
    "MariaDBDsn": MariaDBDsn,
    "ClickHouseDsn": ClickHouseDsn,
    "SnowflakeDsn": SnowflakeDsn,
    "EmailStr": EmailStr,
    "NameEmail": NameEmail,
    "IPvAnyAddress": IPvAnyAddress,
}

# --- Unified Map ---
PRIMITIVE_TYPE_MAP = {**STANDARD_TYPES, **ASTROPY_TYPES}



# validate_config_with_lines.py
import json
import logging
import os
import sys
import tomllib
import traceback
from collections.abc import Callable
from enum import Enum
from io import StringIO
from pathlib import Path
from typing import Any, Optional, TextIO, cast

from pydantic import (
    BaseModel,
    Field,
    ValidationError,
    create_model,
)
from ruamel.yaml import YAML, YAMLError

from yasl.cache import YaslRegistry
from yasl.primitives import PRIMITIVE_TYPE_MAP
from yasl.pydantic_types import Enumeration, TypeDef, YASLBaseModel, YaslRoot
from yasl.validators import property_validator_factory, type_validator_factory


# --- Logging Setup ---
class YamlFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        yaml = YAML()
        log_dict = {
            "level": record.levelname,
            "name": record.name,
            "message": record.getMessage(),
            "time": self.formatTime(record, self.datefmt),
        }
        stream = StringIO()
        log_list = []
        log_list.append(log_dict)
        yaml.dump(log_list, stream)
        return stream.getvalue().strip()


class JsonFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        log_dict = {
            "level": record.levelname,
            "name": record.name,
            "message": record.getMessage(),
            "time": self.formatTime(record, self.datefmt),
        }
        return json.dumps(log_dict)


def setup_logging(
    disable: bool,
    verbose: bool,
    quiet: bool,
    output: str,
    stream: StringIO | TextIO = sys.stdout,
):
    logger = logging.getLogger()
    logger.handlers.clear()
    if disable:
        logger.disabled = True
        return
    if verbose:
        level = logging.DEBUG
    elif quiet:
        level = logging.ERROR
    else:
        level = logging.INFO
    logger.setLevel(level)
    handler = logging.StreamHandler(stream)
    if output == "json":
        handler.setFormatter(JsonFormatter())
    elif output == "yaml":
        handler.setFormatter(YamlFormatter())
    else:
        handler.setFormatter(logging.Formatter("%(message)s"))
    logger.addHandler(handler)


# --- YASL Models and Validation Logic ---


def yasl_version() -> str:
    try:
        pyproject_path = os.path.join(os.path.dirname(__file__), "../../pyproject.toml")
        with open(pyproject_path, "rb") as f:
            pyproject = tomllib.load(f)
        return pyproject["project"]["version"]
    except Exception:
        # fallback to old version if pyproject.toml is missing or malformed
        return "Unknown due to internal error reading pyproject.toml"


def yasl_eval(
    yasl_schema: str,
    yaml_data: str,
    model_name: str | None = None,
    disable_log: bool = False,
    quiet_log: bool = False,
    verbose_log: bool = False,
    output: str = "text",
    log_stream: StringIO | TextIO = sys.stdout,
) -> list[BaseModel] | None:
    """
    Evaluate YAML data against a YASL schema.

    Args:
        yasl_schema (str): Path to the YASL schema file or directory.
        yaml_data (str): Path to the YAML data file or directory.
        model_name (str, optional): Specific model name to use for validation. If not provided, the model will be auto-detected.
        disable_log (bool): If True, disables all logging output.
        quiet_log (bool): If True, suppresses all output except for errors.
        verbose_log (bool): If True, enables verbose logging output.
        output (str): Output format for logs. Options are 'text', 'json', or 'yaml'. Default is 'text'.
        log_stream (StringIO): Stream to which logs will be written. Default is sys.stdout.

    Returns:
        Optional[List[BaseModel]]: List of validated Pydantic models if validation is successful, None otherwise.
    """

    setup_logging(
        disable=disable_log,
        verbose=verbose_log,
        quiet=quiet_log,
        output=output,
        stream=log_stream,
    )
    log = logging.getLogger("yasl")
    log.debug(f"YASL Version - {yasl_version()}")
    log.debug(f"YASL Schema - {yasl_schema}")
    log.debug(f"YAML Data - {yaml_data}")

    registry = YaslRegistry()

    yasl_files = []
    if Path(yasl_schema).is_dir():
        for p in Path(yasl_schema).rglob("*.yasl"):
            yasl_files.append(p)
        if not yasl_files:
            log.error(f"❌ No .yasl files found in directory '{yasl_schema}'")
            registry.clear_caches()
            return None
        log.debug(f"Found {len(yasl_files)} .yasl files in directory '{yasl_schema}'")
    else:
        if not Path(yasl_schema).exists():
            log.error(f"❌ YASL schema file '{yasl_schema}' not found")
            registry.clear_caches()
            return None
        yasl_files.append(Path(yasl_schema))

    yaml_files = []
    if Path(yaml_data).is_dir():
        for p in Path(yaml_data).rglob("*.yaml"):
            yaml_files.append(p)
        if not yaml_files:
            log.error(f"❌ No .yaml files found in directory '{yaml_data}'")
            registry.clear_caches()
            return None
        log.debug(f"Found {len(yaml_files)} .yaml files in directory '{yaml_data}'")
    else:
        if not Path(yaml_data).exists():
            log.error(f"❌ YAML data file '{yaml_data}' not found")
            registry.clear_caches()
            return None
        yaml_files.append(Path(yaml_data))

    yasl_results = []
    for yasl_file in yasl_files:
        yasl = load_and_validate_yasl_with_lines(yasl_file)
        if yasl is None:
            log.error("❌ YASL schema validation failed. Exiting.")
            registry.clear_caches()
            return None
        yasl_results.extend(yasl)

    results = []

    for yaml_file in yaml_files:
        results = load_and_validate_data_with_lines(model_name, yaml_file)

        if not results or len(results) == 0:
            log.error(
                f"❌ Validation failed. Unable to validate data in YAML file {yaml_file}."
            )
            registry.clear_caches()
            return None

    registry.clear_caches()
    return results


def gen_enum_from_enumerations(namespace: str, enum_defs: dict[str, Enumeration]):
    """
    Dynamically generate a Python Enum class from an Enumeration instance.
    Each value in the Enumeration becomes a member of the Enum.
    """
    registry = YaslRegistry()
    for enum_name, enum_def in enum_defs.items():
        if registry.get_enum(enum_name, namespace) is not None:
            raise ValueError(f"Enumeration '{namespace}.{enum_name}' already exists.")
        enum_members = {value: value for value in enum_def.values}
        enum_cls = Enum(enum_name, enum_members)
        enum_cls.__module__ = namespace
        registry.register_enum(enum_name, enum_cls, namespace)


def gen_pydantic_type_models(namespace: str, type_defs: dict[str, TypeDef]):
    """
    Dynamically generate Pydantic model classes from a list of TypeDef instances.
    Each property in the TypeDef becomes a field in the generated model.
    """
    registry = YaslRegistry()
    for typedef_name, type_def in type_defs.items():
        if registry.get_type(typedef_name, namespace) is not None:
            raise ValueError(
                f"Type definition '{namespace}.{typedef_name}' already exists."
            )
        fields: dict[str, tuple] = {}
        validators: dict[str, Callable] = {}
        for prop_name, prop in type_def.properties.items():
            # Determine type annotation for the property
            # For now, map basic types; extend as needed for complex types
            type_map = PRIMITIVE_TYPE_MAP
            type_lookup = prop.type
            type_lookup_namespace = None
            is_list = False
            is_map = False
            key = None
            if type_lookup.endswith("[]"):
                type_lookup = prop.type[:-2]
                is_list = True

            if (
                "ref[" not in type_lookup
                and "map[" not in type_lookup
                and "." in type_lookup
            ):
                # likely a ref to another type or enum
                parts = type_lookup.split(".")
                type_lookup = parts[-1]
                type_lookup_namespace = ".".join(parts[:-1])

            if type_lookup in type_map:
                py_type = type_map[type_lookup]
            elif (
                registry.get_enum(type_lookup, type_lookup_namespace, namespace)
                is not None
            ):
                py_type = registry.get_enum(
                    type_lookup, type_lookup_namespace, namespace
                )
            elif (
                registry.get_type(type_lookup, type_lookup_namespace, namespace)
                is not None
            ):
                py_type = registry.get_type(
                    type_lookup, type_lookup_namespace, namespace
                )
            elif type_lookup.startswith("map[") and type_lookup.endswith("]"):
                is_map = True
                key, value = type_lookup[4:-1].split(",", 1)

                key_type_lookup = key.strip()
                key_type_lookup_namespace = None
                if "." in key_type_lookup:
                    key_type_lookup_namespace, key_type_lookup = key_type_lookup.rsplit(
                        ".", 1
                    )
                # make sure map key is a known type

                if key in ["str", "string"]:
                    key = str
                elif key == "int":
                    key = int
                elif (
                    registry.get_enum(
                        key_type_lookup, key_type_lookup_namespace, namespace
                    )
                    is not None
                ):
                    key = registry.get_enum(
                        key_type_lookup, key_type_lookup_namespace, namespace
                    )
                else:
                    acceptable_keys = [
                        "str",
                        "string",
                        "int",
                    ] + registry.get_enum_names()
                    raise ValueError(
                        f"Map key type '{key}' for property '{prop_name}' must be one of {acceptable_keys}."
                    )

                value_type_lookup = value.strip()
                value_type_lookup_namespace = None
                # if map value is a list, handle that
                map_value_is_list = False

                if value_type_lookup.endswith("[]"):
                    value_type_lookup = value_type_lookup[:-2]
                    map_value_is_list = True

                # make sure map value is a known type
                if "." in value_type_lookup:
                    value_type_lookup_namespace, value_type_lookup = (
                        value_type_lookup.rsplit(".", 1)
                    )
                if value_type_lookup in type_map:
                    py_type = type_map[value_type_lookup]
                elif (
                    registry.get_enum(
                        value_type_lookup, value_type_lookup_namespace, namespace
                    )
                    is not None
                ):
                    py_type = registry.get_enum(
                        value_type_lookup, value_type_lookup_namespace, namespace
                    )
                elif (
                    registry.get_type(
                        value_type_lookup, value_type_lookup_namespace, namespace
                    )
                    is not None
                ):
                    py_type = registry.get_type(
                        value_type_lookup, value_type_lookup_namespace, namespace
                    )
                else:
                    raise ValueError(
                        f"Unknown map value type '{value_type_lookup}' for property '{prop_name}'"
                    )

                # wrap in list if needed
                if map_value_is_list:
                    py_type = list[py_type]
            elif type_lookup.startswith("ref[") and type_lookup.endswith("]"):
                ref_target = type_lookup[4:-1]
                if "." not in ref_target:
                    raise ValueError(
                        f"Reference '{ref_target}' for property '{prop_name}' must be in the format TypeName.PropertyName or Namespace.TypeName.PropertyName"
                    )
                ref_type_name, property_name = ref_target.rsplit(".", 1)
                ref_type_namespace = None
                if "." in ref_type_name:
                    ref_type_namespace, ref_type_name = ref_type_name.rsplit(".", 1)
                target_type = registry.get_type(
                    ref_type_name, ref_type_namespace, namespace
                )
                if not target_type:
                    raise ValueError(
                        f"Referenced type '{ref_type_name}' for property '{prop_name}' not found in type definitions"
                    )
                else:
                    target_prop = next(
                        (
                            p
                            for p_name, p in type_defs[
                                target_type.__name__
                            ].properties.items()
                            if p_name == property_name
                        ),
                        None,
                    )
                    if not target_prop:
                        raise ValueError(
                            f"Referenced property '{property_name}' in type '{ref_type_name}' not found for property '{prop_name}'"
                        )
                    else:
                        if not target_prop.unique:
                            raise ValueError(
                                f"Referenced property '{ref_type_name}.{property_name}' must be unique to be used as a reference for property '{typedef_name}.{prop_name}'"
                            )
                        elif target_prop.type not in type_map:
                            raise ValueError(
                                f"Referenced property '{ref_type_name}.{property_name}' must be a primitive type to be used as a reference for property '{typedef_name}.{prop_name}'"
                            )
                        else:
                            py_type = type_map[target_prop.type]
            else:
                raise ValueError(
                    f"Unknown type '{prop.type}' for property '{prop_name}'"
                )

            if is_list and is_map:
                raise ValueError(
                    f"Property '{prop_name}' cannot be both a list and a map"
                )

            if is_list:
                py_type = list[py_type]

            if is_map:
                py_type = dict[key, py_type]

            # Handle presence
            is_required = False
            if prop.presence == "required":
                is_required = True
            elif prop.presence == "preferred":
                is_required = False
            elif prop.presence == "optional":
                is_required = False
            else:
                # Default to optional if None (though default is "optional" in model)
                is_required = False

            if not is_required:
                py_type = Optional[py_type]

            default = (
                prop.default
                if prop.default is not None
                else (None if not is_required else ...)
            )
            fields[prop_name] = (py_type, default)
            validators[f"{prop_name}__validator"] = property_validator_factory(
                typedef_name, namespace, type_def, prop_name, prop
            )

        # Add hidden field for YAML line number
        fields["yaml_line"] = (Optional[int], Field(default=None, exclude=True))

        validators["__validate__"] = type_validator_factory(type_def)
        model = create_model(  # type: ignore
            typedef_name,
            __base__=YASLBaseModel,
            __module__=namespace,
            __validators__=validators,
            __config__={"extra": "forbid"},
            **fields,  # type: ignore
        )
        # Store the generated model in the global registry
        registry.register_type(typedef_name, model, namespace)


# --- Helper function to find the line number ---
def get_line_for_error(data: Any, loc: tuple[str | int, ...]) -> int | None:
    """Traverse the ruamel.yaml data to find the line number for an error location."""
    current_data = data
    try:
        for key in loc:
            current_data = current_data[key]
        # .lc is the line/column accessor in ruamel.yaml
        return current_data.lc.line + 1
    except (KeyError, IndexError, AttributeError):
        # Fallback if we can't find the exact key (e.g., for a missing key)
        # We can try to get the line of the parent object.
        parent_data = data
        for key in loc[:-1]:
            parent_data = parent_data[key]
        try:
            return parent_data.lc.line + 1
        except AttributeError:
            return None


# --- Main schema validation logic ---
def load_and_validate_yasl_with_lines(path: str) -> list[YaslRoot] | None:
    log = logging.getLogger("yasl")
    log.debug(f"--- Attempting to validate schema '{path}' ---")
    data = None
    try:
        results = []
        yaml_loader = YAML(typ="rt")
        docs = []
        with open(path) as f:
            docs.extend(yaml_loader.load_all(f))

        for data in docs:
            yasl = YaslRoot(**data)
            if yasl is None:
                raise ValueError("Failed to parse YASL schema from data {data}")
            if yasl.imports is not None:
                for imp in yasl.imports:
                    imp_path = imp
                    if not Path(imp_path).exists():
                        # try relative to current schema file
                        imp_path = Path(path).parent / imp
                        if not imp_path.exists():
                            raise FileNotFoundError(f"Import file '{imp}' not found")
                        imp_path = imp_path.as_posix()
                    log.debug(
                        f"Importing additional schema '{imp}' - resolved to '{imp_path}'"
                    )
                    imported_yasl = load_and_validate_yasl_with_lines(imp_path)
                    if not imported_yasl:
                        raise ValueError(f"Failed to import YASL schema from '{imp}'")
            if yasl.metadata is not None:
                log.debug(f"YASL Metadata: {yasl.metadata}")
            if yasl.definitions is not None:
                for namespace, yasl_item in yasl.definitions.items():
                    # generate enums first so enum map keys are known when generating types
                    if yasl_item.enums is not None:
                        gen_enum_from_enumerations(namespace, yasl_item.enums)
                for namespace, yasl_item in yasl.definitions.items():
                    if yasl_item.types is not None:
                        gen_pydantic_type_models(namespace, yasl_item.types)
            results.append(yasl)
        if not results or len(results) == 0:
            log.error(f"❌ No YASL schema definitions found in '{path}'")
            return None
        log.debug("✅ YASL schema validation successful!")
        return results
    except FileNotFoundError:
        log.error(f"❌ Error - YASL schema file not found at '{path}'")
        return None
    except SyntaxError as e:
        log.error(f"❌ Error - Syntax error in YASL schema file '{path}'\n  - {e}")
        return None
    except YAMLError as e:
        log.error(f"❌ Error - YAML error while parsing YASL schema '{path}'\n  - {e}")
        return None
    except ValidationError as e:
        log.error(
            f"❌ YASL schema validation of {path} failed with {len(e.errors())} error(s):"
        )
        for error in e.errors():
            line = get_line_for_error(data, error["loc"])
            path_str = " -> ".join(map(str, error["loc"]))
            if line:
                log.error(f"  - Line {line}: '{path_str}' -> {error['msg']}")
            else:
                log.error(f"  - Location '{path_str}' -> {error['msg']}")
        return None
    except Exception as e:
        log.error(f"❌ An schema error occurred processing '{path}' - {type(e)} - {e}")
        return None


# --- Main data validation logic ---
def load_and_validate_data_with_lines(model_name: str | None, path: str) -> Any:
    log = logging.getLogger("yasl")
    log.debug(f"--- Attempting to validate data '{path}' ---")
    docs = []
    data = None
    try:
        yaml_loader = YAML(typ="rt")
        with open(path) as f:
            docs.extend(yaml_loader.load_all(f))

    except FileNotFoundError:
        log.error(f"❌ Error - File not found at '{path}'")
        return None
    except SyntaxError as e:
        log.error(f"❌ Error - Syntax error in data file '{path}'\n  - {e}")
        return None
    except YAMLError as e:
        log.error(f"❌ Error - YAML error while parsing data '{path}'\n  - {e}")
        return None
    except ValueError as e:
        log.error(f"❌ Error - value error while parsing data '{path}'\n  - {e}")
        return None
    except Exception as e:
        log.error(f"❌ An unexpected error occurred - {type(e)} - {e}")
        traceback.print_exc()
        return None
    try:
        results = []
        registry = YaslRegistry()
        for data in docs:
            candidate_model_names: list[tuple[str, str | None]] = []
            if model_name is None:
                root_keys: list[str] = list(data.keys())
                log.debug(f"Auto-detecting schema for YAML root keys in '{path}'")
                yasl_result = registry.get_types()
                for type_id, type_def in yasl_result.items() or []:
                    type_name, type_namespace = type_id
                    type_def_root_keys: list[str] = list(type_def.model_fields.keys())
                    if all(k in type_def_root_keys for k in root_keys):
                        log.debug(
                            f"Auto-detected root model '{type_name}' for YAML file '{path}'"
                        )
                        candidate_model_names.append((type_name, type_namespace))
            else:
                registry_item = registry.get_type(model_name)
                if registry_item:
                    candidate_model_names.append(
                        (model_name, registry_item.__module__ or None)
                    )
                else:
                    candidate_model_names.append((model_name, None))

            log.debug(
                f"Identified candidate model names for '{path}' - {candidate_model_names}"
            )

            for schema_name, schema_namespace in candidate_model_names:
                if registry.get_type(schema_name, schema_namespace) is None:
                    continue
                model = registry.get_type(schema_name, schema_namespace)
                log.debug(
                    f"Using schema '{schema_name}' for data validation of {path}."
                )

                # Inject line number if available
                if hasattr(data, "lc") and hasattr(data.lc, "line"):
                    data["yaml_line"] = data.lc.line + 1

                if model is not None:
                    result = cast(type[BaseModel], model)(**data)  # type: ignore
                    if result is not None:
                        results.append(result)
                        break
                    else:
                        log.debug(
                            f"Data in '{path}' did not validate against schema '{schema_name}'."
                        )

        if not results or len(results) == 0:
            log.error(f"❌ No valid schema found to validate data in '{path}'")
            return None
        log.info(f"✅ YAML '{path}' data validation successful!")
        return results
    except ValidationError as e:
        log.error(f"❌ Validation failed with {len(e.errors())} error(s):")
        for error in e.errors():
            line = get_line_for_error(data, error["loc"])
            path_str = " -> ".join(map(str, error["loc"]))
            if line:
                log.error(f"  - Line {line} - '{path_str}' -> {error['msg']}")
            else:
                log.error(f"  - Location '{path_str}' -> {error['msg']}")
        return None
    except SyntaxError as e:
        log.error(
            f"❌ SyntaxError in file '{path}' "
            f"at line {getattr(e, 'lineno', '?')}, offset {getattr(e, 'offset', '?')} - {getattr(e, 'msg', str(e))}"
        )
        if hasattr(e, "text") and e.text:
            log.error(f"  > {e.text.strip()}")
        return None
    except Exception as e:
        log.error(f"❌ An unexpected error occurred - {type(e)} - {e}")
        traceback.print_exc()
        return None



import datetime
import logging
import re
from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

import markdown_it
import requests
from pydantic import field_validator, model_validator

from yasl.cache import YaslRegistry
from yasl.pydantic_types import IfThen, Property, TypeDef


def unique_value_validator(
    cls,
    value: Any,
    type_name: str,
    property_name: str,
    type_namespace: str | None = None,
):
    registry = YaslRegistry()
    registry.register_unique_value(type_name, property_name, value, type_namespace)
    return value


# list validation
def list_min_validator(cls, value: list[Any], bound: int):
    if len(value) < bound:
        raise ValueError(f"List must contain at least {bound} items")
    return value


def list_max_validator(cls, value: list[Any], bound: int):
    if len(value) > bound:
        raise ValueError(f"List must contain at most {bound} items")
    return value


# numeric validation
def gt_validator(cls, value, bound):
    if value <= bound:
        raise ValueError(f"Value must be greater than {bound}")
    return value


def ge_validator(cls, value, bound):
    if value < bound:
        raise ValueError(f"Value must be greater than or equal to {bound}")
    return value


def lt_validator(cls, value, bound):
    if value >= bound:
        raise ValueError(f"Value must be less than {bound}")
    return value


def le_validator(cls, value, bound):
    if value > bound:
        raise ValueError(f"Value must be less than or equal to {bound}")
    return value


def exclude_validator(cls, value, excluded_values):
    if value in excluded_values:
        raise ValueError(f"Value must not be one of {excluded_values}")
    return value


def multiple_of_validator(cls, value, bound):
    if value % bound != 0:
        raise ValueError(f"Value must be a multiple of {bound}")
    return value


# string validation
def str_min_validator(cls, value: str, min_length: int):
    if len(value) < min_length:
        raise ValueError(f"String must be at least {min_length} characters long")
    return value


def str_max_validator(cls, value: str, max_length: int):
    if len(value) > max_length:
        raise ValueError(f"String must be at most {max_length} characters long")
    return value


def str_regex_validator(cls, value: str, regex: str):
    if not re.fullmatch(regex, value):
        raise ValueError(f"Value '{value}' does not match pattern '{regex}'")
    return value


# date validators
def date_before_validator(
    cls,
    value: datetime.datetime | datetime.date | datetime.time,
    before: datetime.datetime | datetime.date | datetime.time,
):
    if not isinstance(value, type(before)):
        raise ValueError(
            f"Type of value '{type(value)}' does not match type of 'before' '{type(before)}'"
        )
    if value >= before:  # type: ignore
        raise ValueError(f"Date '{value}' must be before '{before}'")
    return value


def date_after_validator(
    cls,
    value: datetime.datetime | datetime.date | datetime.time,
    after: datetime.datetime | datetime.date | datetime.time,
):
    if not isinstance(value, type(after)):
        raise ValueError(
            f"Type of value '{type(value)}' does not match type of 'after' '{type(after)}'"
        )
    if value <= after:  # type: ignore
        raise ValueError(f"Date '{value}' must be after '{after}'")
    return value


# url validators
def url_base_validator(cls, value: str, url_base: str):
    parsed = urlparse(value)
    if url_base != parsed.netloc:
        raise ValueError(f"URL '{value}' does not have a base of '{url_base}'")
    return value


def url_protocol_validator(cls, value: str, protocols: list[str]):
    parsed = urlparse(value)
    if parsed.scheme and parsed.scheme not in protocols:
        raise ValueError(f"URL '{value}' must use one of the protocols {protocols}")
    return value


def url_reachable_valiator(cls, value: str, reachable: bool):
    if reachable:
        try:
            response = requests.head(value, allow_redirects=True, timeout=3)
            if response.status_code >= 400:
                raise ValueError(
                    f"URL '{value}' is not reachable (status {response.status_code})"
                )
        except requests.RequestException as e:
            raise ValueError(f"URL '{value}' is not reachable: {e}") from e
    return value


def is_dir_validator(cls, value: str, is_dir: bool):
    path = Path(value)
    # Check if the path looks like a directory (ends with separator or no suffix)
    if not (path.suffix == "" or value.endswith(("/", "\\"))):
        raise ValueError(f"Path '{value}' must be a directory")
    return value


def is_file_validator(cls, value: str, is_file: bool):
    path = Path(value)
    # Check if the path looks like a file (does not end with a separator, has a suffix)
    if not (path.suffix != "" and not value.endswith(("/", "\\"))):
        raise ValueError(f"Path '{value}' must be a file")
    return value


def file_ext_validator(cls, value: str, extensions: list[str]):
    path = Path(value)
    # First, ensure it's a file
    is_file_validator(cls, value, True)
    # Check if the file has one of the allowed extensions
    if extensions and path.suffix not in [
        f".{ext}" if not ext.startswith(".") else ext for ext in extensions
    ]:
        raise ValueError(
            f"File '{value}' must have one of the following extensions: {extensions}"
        )
    return value


def path_exists_validator(cls, value: str, exists: bool):
    path = Path(value)
    # Check if the path exists on the filesystem
    if exists and not path.exists():
        raise ValueError(f"Path '{value}' must exist on the filesystem")
    return value


# ref validators
def ref_exists_validator(cls, value: Any, target: str):
    type_name, property_name = target.rsplit(".", 1)
    type_namespace = None
    if type_name and "." in type_name:
        type_namespace, type_name = type_name.rsplit(".", 1)

    registry = YaslRegistry()
    if not registry.unique_value_exists(
        type_name, property_name, value, type_namespace
    ):
        raise ValueError(
            f"Referenced value '{value}' does not exist for 'ref[{target}]"
        )

    return value


# any validator
def any_of_validator(cls, value: Any, any_of: list[str]):
    for t in any_of:
        if t.endswith("[]"):
            elem_type = t[:-2]
            if isinstance(value, list) and all(
                isinstance(v, eval(elem_type)) for v in value
            ):
                return value
        else:
            if isinstance(value, eval(t)):
                return value
    raise ValueError(f"Value '{value}' must be one of {any_of}")


# enum validator
def enum_validator(cls, value: Any, values: list[str]):
    str_value = str(value)
    if str_value.split(".")[-1] not in values:
        raise ValueError(f"Value '{value}' must be one of {values}")
    return value


# map validator
def map_validator(
    cls,
    value: dict[Any, Any],
    key_type: str,
    value_type: str,
    any_of: list[str] | None = None,
):
    # validate key type is str, int, or an enumation
    registry = YaslRegistry()
    enum_namespace = None
    enum_name = key_type
    if "." in key_type:
        enum_namespace, enum_name = key_type.rsplit(".", 1)
    if (
        key_type not in ["str", "int"]
        and registry.get_enum(enum_name, enum_namespace) is None
    ):
        raise ValueError(f"Map key type '{key_type}' is not supported")
    # validate value type of any is allowed by constraints
    if value_type == "any" and any_of is not None:
        if not any(isinstance(v, eval(t)) for t in any_of for v in value.values()):
            raise ValueError(f"Map values must be one of {any_of}")
    return value


# markdown validator
def markdown_validator(cls, value: str):
    try:
        md = markdown_it.MarkdownIt()
        tokens = md.parse(value)
        if not tokens:
            raise ValueError("Markdown content is empty or invalid.")
        return value
    except Exception as e:
        raise ValueError("Markdown content is not valid.") from e


def property_validator_factory(
    typedef_name: str,
    type_namespace: str,
    type_def: TypeDef,
    property_name: str,
    property: Property,
) -> Callable:
    validators = []
    # list validators
    if property.list_min is not None:
        validators.append(partial(list_min_validator, bound=property.list_min))
    if property.list_max is not None:
        validators.append(partial(list_max_validator, bound=property.list_max))

    # unique validator
    if property.unique:
        if "." in property_name:
            property_namespace, property_name = property_name.rsplit(".", 1)
        else:
            property_namespace = type_namespace
        validators.append(
            partial(
                unique_value_validator,
                type_name=typedef_name,
                property_name=property_name,
                type_namespace=property_namespace,
            )
        )

    # numeric validators
    if property.gt is not None:
        validators.append(partial(gt_validator, bound=property.gt))
    if property.ge is not None:
        validators.append(partial(ge_validator, bound=property.ge))
    if property.lt is not None:
        validators.append(partial(lt_validator, bound=property.lt))
    if property.le is not None:
        validators.append(partial(le_validator, bound=property.le))
    if property.exclude is not None:
        validators.append(partial(exclude_validator, excluded_values=property.exclude))
    if property.multiple_of is not None:
        validators.append(partial(multiple_of_validator, bound=property.multiple_of))

    # string validators
    if property.str_min is not None:
        validators.append(partial(str_min_validator, min_length=property.str_min))
    if property.str_max is not None:
        validators.append(partial(str_max_validator, max_length=property.str_max))
    if property.str_regex is not None:
        validators.append(partial(str_regex_validator, regex=property.str_regex))

    # date validators
    if property.before is not None:
        validators.append(partial(date_before_validator, before=property.before))
    if property.after is not None:
        validators.append(partial(date_after_validator, after=property.after))

    # path validators
    if property.path_exists is not None:
        validators.append(partial(path_exists_validator, exists=property.path_exists))
    if property.is_dir is not None:
        validators.append(partial(is_dir_validator, is_dir=property.is_dir))
    if property.is_file is not None:
        validators.append(partial(is_file_validator, is_file=property.is_file))
    if property.file_ext is not None:
        validators.append(partial(file_ext_validator, extensions=property.file_ext))

    # uri validators
    if property.url_base is not None:
        validators.append(partial(url_base_validator, url_base=property.url_base))
    if property.url_protocols is not None:
        validators.append(
            partial(url_protocol_validator, protocols=property.url_protocols)
        )
    if property.url_reachable is not None:
        validators.append(
            partial(url_reachable_valiator, reachable=property.url_reachable)
        )

    # any validator
    if property.any_of is not None and not property.type.startswith("map[]"):
        validators.append(partial(any_of_validator, any_of=property.any_of))

    # ref validators (always validate references)
    if property.type.startswith("ref[") and (
        property.no_ref_check is None or property.no_ref_check is False
    ):
        validators.append(partial(ref_exists_validator, target=property.type[4:-1]))

    # enum validators
    registry = YaslRegistry()
    enum_names: list[tuple[str, str | None]] = registry.get_enum_names()

    if property.type in dict(enum_names).keys():
        enum_type = registry.get_enum(property.type, type_def.namespace)
        if enum_type is None:
            raise ValueError(
                f"Enum type '{property.type}' not found for property '{property_name}' in type '{typedef_name}'"
            )
        validators.append(partial(enum_validator, values=[e.value for e in enum_type]))  # type: ignore

    # map validators
    if property.type.startswith("map["):
        # extract key and value types
        map_types = property.type[4:-1]
        if "," not in map_types:
            raise ValueError(f"Map type '{property.type}' is not valid")
        key_type, value_type = map(str.strip, map_types.split(",", 1))
        validators.append(
            partial(
                map_validator,
                key_type=key_type,
                value_type=value_type,
                any_of=property.any_of,
            )
        )

    # markdown validator
    if property.type == "markdown":
        validators.append(markdown_validator)

    def multi_validator(cls, value):
        for validator in validators:
            value = validator(cls, value)
        return value

    return field_validator(property_name)(multi_validator)


# type validators
def only_one_validator(cls, values: dict[str, Any], fields: list[str]):
    data_keys = [key for key, val in values.model_dump().items() if val is not None]  # type: ignore
    if sum(1 for field in fields if field in data_keys) != 1:
        raise ValueError(f"Exactly one of {fields} must be present")
    return values


def at_least_one_validator(cls, values: dict[str, Any], fields: list[str]):
    data_keys = [key for key, val in values.model_dump().items() if val is not None]  # type: ignore
    if sum(1 for field in fields if field in data_keys) < 1:
        raise ValueError(f"At least one of {fields} must be present")
    return values


def if_then_validator(cls, values: dict[str, Any], if_then: IfThen):
    eval_field = if_then.eval
    eval_value = if_then.value
    present_fields = if_then.present or []
    absent_fields = if_then.absent or []
    values_dict = values.model_dump()  # type: ignore
    if eval_field in values_dict:
        eval_value_type = type(values_dict[eval_field])
        typed_eval_value = [eval_value_type(v) for v in eval_value]
        if values_dict[eval_field] in typed_eval_value:
            for field in present_fields:
                if field not in values_dict or values_dict[field] is None:
                    raise ValueError(
                        f"Field '{field}' must be present when '{eval_field}' is in {eval_value}"
                    )
            for field in absent_fields:
                if field in values_dict and values_dict[field] is not None:
                    raise ValueError(
                        f"Field '{field}' must be absent when '{eval_field}' is in {eval_value}"
                    )
    return values


def preferred_presence_validator(cls, values: Any, properties: dict[str, Property]):
    log = logging.getLogger("yasl")
    values_dict = values.model_dump()

    # Access the hidden yaml_line field if it exists
    line_info = ""
    if hasattr(values, "yaml_line") and values.yaml_line is not None:
        line_info = f" at line {values.yaml_line}"

    for prop_name, prop in properties.items():
        if prop.presence == "preferred":
            if prop_name not in values_dict or values_dict[prop_name] is None:
                log.warning(
                    f"⚠️  Warning: Preferred property '{prop_name}' is missing{line_info}"
                )
    return values


def type_validator_factory(model: TypeDef) -> Callable:
    validators = []
    if model.validators is not None:
        if model.validators.only_one is not None:
            validators.append(
                partial(only_one_validator, fields=model.validators.only_one)
            )
        if model.validators.at_least_one is not None:
            validators.append(
                partial(at_least_one_validator, fields=model.validators.at_least_one)
            )
        if model.validators.if_then is not None:
            for item in model.validators.if_then:
                validators.append(partial(if_then_validator, if_then=item))

    if model.properties:
        validators.append(
            partial(preferred_presence_validator, properties=model.properties)
        )

    @model_validator(mode="after")
    def multi_validator(cls, values):
        for validator in validators:
            values = validator(cls, values)
        return values

    return multi_validator  # type: ignore



import logging
from enum import Enum
from types import MappingProxyType
from typing import Any, Optional

from pydantic import BaseModel


class YaslRegistry:
    """
    Singleton registry for YASL type definitions and enumerations.
    Supports registration and lookup by name and optional namespace.
    """

    _instance: Optional["YaslRegistry"] = None

    def __new__(cls) -> "YaslRegistry":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._init_registry()
        return cls._instance

    def _init_registry(self) -> None:
        self.yasl_type_defs: dict[tuple[str, str | None], BaseModel] = {}
        self.yasl_enumerations: dict[tuple[str, str | None], Enum] = {}
        self.unique_values_store: dict[tuple[str, str | None], dict[str, set]] = {}

    def register_type(self, name: str, type_def: BaseModel, namespace: str) -> None:
        key = (name, namespace)
        log = logging.getLogger("yasl")
        if key in self.yasl_type_defs:
            raise ValueError(f"Type '{name}' already exists in namespace '{namespace}'")
        self.yasl_type_defs[key] = type_def
        log.debug(f"Registered type '{name}' in namespace '{namespace}'")

    def get_types(self) -> MappingProxyType[tuple[str, str | None], BaseModel]:
        # Return a read-only view of the registered types
        return MappingProxyType(self.yasl_type_defs)

    def get_type(
        self,
        name: str,
        namespace: str | None = None,
        default_namespace: str | None = None,
    ) -> BaseModel | None:
        log = logging.getLogger("yasl")
        log.debug(
            f"Looking up type '{name}' in namespace '{namespace}' with default namespace '{default_namespace}'"
        )
        if namespace is not None:
            key = (name, namespace)
            if key in self.yasl_type_defs:
                return self.yasl_type_defs[key]
            return None
        matches = [v for (n, ns), v in self.yasl_type_defs.items() if n == name]
        if not matches:
            return None
        if len(matches) == 1:
            return matches[0]
        elif default_namespace is not None:
            log.debug(
                f"Trying default namespace '{default_namespace}' for type '{name}'"
            )
            key = (name, default_namespace)
            if key in self.yasl_type_defs:
                return self.yasl_type_defs[key]
        raise ValueError(
            f"Ambiguous type name '{name}': found in multiple namespaces {matches}. Specify a namespace."
        )

    def register_enum(self, name: str, enum_def: Enum, namespace: str) -> None:
        log = logging.getLogger("yasl")
        key = (name, namespace)
        if key in self.yasl_enumerations:
            raise ValueError(f"Enum '{name}' already exists in namespace '{namespace}'")
        self.yasl_enumerations[key] = enum_def
        log.debug(f"Registered enum '{name}' in namespace '{namespace}'")

    def get_enum_names(self) -> list[tuple[str, str | None]]:
        return [(n, ns) for (n, ns) in self.yasl_enumerations.keys()]

    def get_enum(
        self,
        name: str,
        namespace: str | None = None,
        default_namespace: str | None = None,
    ) -> Enum | None:
        log = logging.getLogger("yasl")
        log.debug(
            f"Looking up enum '{name}' in namespace '{namespace}' with default namespace '{default_namespace}'"
        )
        if namespace is not None:
            key = (name, namespace)
            if key in self.yasl_enumerations:
                return self.yasl_enumerations[key]
            return None
        matches = [v for (n, ns), v in self.yasl_enumerations.items() if n == name]
        if not matches:
            return None
        if len(matches) == 1:
            return matches[0]
        elif default_namespace is not None:
            key = (name, default_namespace)
            if key in self.yasl_enumerations:
                return self.yasl_enumerations[key]
        raise ValueError(
            f"Ambiguous enum name '{name}': found in multiple namespaces {matches}. Specify a namespace."
        )

    def register_unique_value(
        self,
        type_name: str,
        property_name: str,
        value: Any,
        type_namespace: str | None = None,
    ) -> None:
        if (type_name, type_namespace) not in self.unique_values_store:
            self.unique_values_store[type_name, type_namespace] = {}
        if property_name not in self.unique_values_store[type_name, type_namespace]:
            self.unique_values_store[type_name, type_namespace][property_name] = set()
        if value in self.unique_values_store[type_name, type_namespace][property_name]:
            raise ValueError(
                f"Duplicate unique value '{value}' for property '{property_name}' in type '{type_name}'"
            )
        self.unique_values_store[type_name, type_namespace][property_name].add(value)

    def unique_value_exists(
        self,
        type_name: str,
        property_name: str,
        value: Any,
        type_namespace: str | None = None,
    ) -> bool:
        if type_namespace is None:
            matches = [
                (tn, ns)
                for (tn, ns) in self.unique_values_store.keys()
                if tn == type_name
            ]
            if not matches:
                return False
            if len(matches) == 1:
                type_namespace = matches[0][1]
            else:
                raise ValueError(
                    f"Ambiguous type name '{type_name}': found in multiple namespaces. Specify a namespace."
                )

        return (
            (type_name, type_namespace) in self.unique_values_store
            and property_name in self.unique_values_store[type_name, type_namespace]
            and value
            in self.unique_values_store[type_name, type_namespace][property_name]
        )

    def clear_caches(self) -> None:
        """Clean up global stores after validation."""
        self.unique_values_store.clear()
        self.yasl_type_defs.clear()
        self.yasl_enumerations.clear()


# Singleton instance
yasl_registry = YaslRegistry()



"""
YAQL CLI main entry point.
"""

import argparse
import sys

from common import advanced_yaml_version


def main():
    parser = argparse.ArgumentParser(
        description="YAQL - YAML Advanced Query Language CLI Tool"
    )
    # Removed --project-name argument; 'param' will be used for project name in 'init'

    parser.add_argument(
        "--version", action="store_true", help="Show version information and exit"
    )
    parser.add_argument(
        "--quiet", action="store_true", help="Suppress output except for errors"
    )
    parser.add_argument("--verbose", action="store_true", help="Enable verbose output")
    parser.add_argument(
        "--output",
        choices=["text", "json", "yaml"],
        default="text",
        help="Set output format (text, json, yaml). Default is text.",
    )

    args = parser.parse_args()

    if args.verbose and args.quiet:
        print("❌ Cannot use both --quiet and --verbose.")
        sys.exit(1)

    if args.version:
        print(f"YAQL version {advanced_yaml_version()}")
        sys.exit(0)

    print(
        "❌ This is a placeholder for the YAQL CLI tool. Functionality to be implemented."
    )
    sys.exit(1)


if __name__ == "__main__":
    main()



import os
import tomllib


def advanced_yaml_version() -> str:
    try:
        pyproject_path = os.path.join(os.path.dirname(__file__), "../../pyproject.toml")
        with open(pyproject_path, "rb") as f:
            pyproject = tomllib.load(f)
        return pyproject["project"]["version"]
    except Exception:
        # fallback to old version if pyproject.toml is missing or malformed
        return "Unknown due to internal error reading pyproject.toml"



"""
YASL CLI main entry point.
"""

import argparse
import sys

from common import advanced_yaml_version
from yasl import yasl_eval


def main():
    parser = argparse.ArgumentParser(
        description="YASL - YAML Advanced Schema Language CLI Tool"
    )
    # Removed --project-name argument; 'param' will be used for project name in 'init'
    parser.add_argument(
        "schema",
        nargs="?",
        help="YASL schema file or directory",
    )
    parser.add_argument(
        "yaml",
        nargs="?",
        help="YAML data file or directory",
    )
    parser.add_argument(
        "model_name",
        nargs="?",
        help="YASL schema type name for the yaml data file (optional)",
    )
    parser.add_argument(
        "--version", action="store_true", help="Show version information and exit"
    )
    parser.add_argument(
        "--quiet", action="store_true", help="Suppress output except for errors"
    )
    parser.add_argument("--verbose", action="store_true", help="Enable verbose output")
    parser.add_argument(
        "--output",
        choices=["text", "json", "yaml"],
        default="text",
        help="Set output format (text, json, yaml). Default is text.",
    )

    args = parser.parse_args()

    if args.verbose and args.quiet:
        print("❌ Cannot use both --quiet and --verbose.")
        sys.exit(1)

    if args.version:
        print(f"YASL version {advanced_yaml_version()}")
        sys.exit(0)

    if not args.schema or not args.yaml:
        print(
            "❌ requires a YASL file, a YAML schema file, and optionally a model name as parameters."
        )
        parser.print_help()
        sys.exit(1)

    yasl = yasl_eval(
        args.schema,
        args.yaml,
        args.model_name,
        disable_log=False,
        quiet_log=args.quiet,
        verbose_log=args.verbose,
        output=args.output,
    )

    if not yasl:
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()



import datetime
from typing import Any, Literal

from pydantic import BaseModel


class YASLBaseModel(BaseModel):
    def __repr__(self) -> str:
        fields = self.model_dump()  # For Pydantic v2; use self.dict() for v1
        return f"{self.__class__.__name__}({fields})"


# --- YASL Pydantic Models ---
class Enumeration(YASLBaseModel):
    # name: str
    description: str | None = None
    values: list[str]

    model_config = {"extra": "forbid"}


class Property(YASLBaseModel):
    # name: str
    type: str
    description: str | None = None
    presence: Literal["required", "preferred", "optional"] | None = "optional"
    unique: bool | None = False
    default: Any | None = None

    # list constraints
    list_min: int | None = None
    list_max: int | None = None

    # numeric constraints
    gt: float | None = None
    ge: float | None = None
    lt: float | None = None
    le: float | None = None
    exclude: list[float] | None = None
    multiple_of: float | None = None
    whole_number: bool | None = False

    # string constraints
    str_min: int | None = None
    str_max: int | None = None
    str_regex: str | None = None

    # date / time constraints
    before: datetime.date | datetime.datetime | datetime.time | None = None
    after: datetime.date | datetime.datetime | datetime.time | None = None

    # path constraints
    path_exists: bool | None = None
    is_dir: bool | None = None
    is_file: bool | None = None
    file_ext: list[str] | None = None

    # url constraints
    url_base: str | None = None
    url_protocols: list[str] | None = None
    url_reachable: bool | None = False

    # any constraints
    any_of: list[str] | None = None

    # ref constraints
    no_ref_check: bool | None = None

    model_config = {"extra": "forbid"}


class IfThen(YASLBaseModel):
    eval: str
    value: list[str]
    present: list[str]
    absent: list[str]

    model_config = {"extra": "forbid"}


class Validator(YASLBaseModel):
    only_one: list[str] | None = None
    at_least_one: list[str] | None = None
    if_then: list[IfThen] | None = None

    model_config = {"extra": "forbid"}


class TypeDef(YASLBaseModel):
    # name: str
    namespace: str | None = None
    description: str | None = None
    properties: dict[str, Property]
    validators: Validator | None = None

    model_config = {"extra": "forbid"}


class YaslItem(YASLBaseModel):
    description: str | None = None
    enums: dict[str, Enumeration] | None = None
    types: dict[str, TypeDef] | None = None
    model_config = {"extra": "forbid"}


class YaslRoot(YASLBaseModel):
    imports: list[str] | None = None
    metadata: dict[str, Any] | None = None
    definitions: dict[str, YaslItem] | None = None
    model_config = {"extra": "forbid"}
